from argparse import Namespace
from daffi_terminals.router.router import Router, WebHandler


def start_router(args: Namespace):
    web_handler = WebHandler(web_host=args.web_host, web_port=args.web_port)
    router = Router(
        rpc_host=args.rpc_host, rpc_port=args.rpc_port,
        ssl_cert=args.ssl_cert, ssl_key=args.ssl_key, web_handler=web_handler
    )
    router.run()



import logging
from argparse import ArgumentParser

logging.basicConfig(level=logging.INFO)


def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers()
    subparsers.required = True
    subparsers.dest = 'command'
    router_parser = subparsers.add_parser("start-router", help="Start router")
    worker_parser = subparsers.add_parser("start-worker", help="Start worker")

    router_parser.add_argument("--rpc-host", help="Router rpc host (Interaction with workers)", required=True)
    router_parser.add_argument("--rpc-port", help="Router rpc port (Interaction with workers", type=int, required=True)
    router_parser.add_argument("--web-host", help="Router web host (Interaction with user)", required=True)
    router_parser.add_argument("--web-port", help="Router web port (Interaction with user", type=int, required=True)
    router_parser.add_argument("--ssl-cert", default=None, help="Path to ssl certificate")
    router_parser.add_argument("--ssl-key", default=None, help="Path to ssl private key")

    worker_parser.add_argument("--name", default=None,
                               help="Worker name. If not provided default name will be generated")
    worker_parser.add_argument("--rpc-host", help="Router host to connect", required=True)
    worker_parser.add_argument("--rpc-port", help="Router port to connect", type=int, required=True)
    worker_parser.add_argument("--ssl-cert", default=None, help="Path to ssl certificate")
    worker_parser.add_argument("--ssl-key", default=None, help="Path to ssl private key")

    args = parser.parse_args()
    if parser.parse_args().command == "start-router":
        from daffi_terminals.router import start_router
        start_router(args)
    else:
        from daffi_terminals.worker import start_worker
        start_worker(args)


if __name__ == '__main__':
    main()



import os
import sys
import fcntl
import tty
import uuid
import struct
import socket
import logging
import traceback
from enum import Enum
from threading import Thread
from multiprocessing import Pipe

from daffi import Global, FG
from daffi.utils import colors
from daffi.registry import Callback, Fetcher
from daffi.settings import BYTES_CHUNK
from daffi.exceptions import RemoteStoppedUnexpectedly, RemoteCallError
from daffi.decorators import local, __body_unknown__

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("worker")


class Command(bytes, Enum):
    STOP = b""
    DATA = b"1"
    RESIZE = b"2"


class RouterFetcher(Fetcher):
    exec_modifier = FG(receiver="TermRouter")

    def __post_init__(self):
        self.id = str(uuid.uuid4())

    @staticmethod
    def write_to_terminal(term_id):
        __body_unknown__(term_id)

    def on_worker_connect(self, process_name: str):
        """
        Return common metadata information about worker.
        Handler is being executed every time on connection (despite of it is first connection or re-connection)
        """
        math_sign = colors.intense_magenta("\uF50E")
        print(f"{math_sign} Happy exploring! {math_sign}")
        host = socket.gethostname()
        mac = ":".join(["{:02x}".format((uuid.getnode() >> ele) & 0xFF) for ele in range(0, 8 * 6, 8)][::-1])
        return host, mac, process_name, self.id

    @local
    def on_worker_connect_cb(self, g: Global, process_name: str):
        # Send initial information about worker to `Router`
        try:
            self.on_worker_connect(process_name=process_name)
        except RemoteCallError as e:
            if "already exists and is active" in e.message:
                logger.error(traceback.format_exc())
                g.stop()
            else:
                raise


class Worker(Callback):

    def __init__(self, router_fetcher: RouterFetcher):
        super().__init__()
        self.router_fetcher = router_fetcher

    def read_from_terminal(self, term_id):
        exeargv = [os.getenv("SHELL", "sh")]
        host_env = os.environ.copy()
        # os.environ["TERM"] = "xterm"
        r, w = Pipe(duplex=False)

        interactive = sys.stdin.isatty()
        if interactive:
            ttyattr = tty.tcgetattr(0)
            # struct winsize 4 unsigned short. see ioctl_tty
            winsize = fcntl.ioctl(0, tty.TIOCGWINSZ, 8 * b" ")

        pid, fdm = os.forkpty()
        if pid == 0:
            # python's forkpty can't pass termios and winsize. do manually
            # see forkpty
            if interactive:
                tty.tcsetattr(0, tty.TCSANOW, ttyattr)
                fcntl.ioctl(0, tty.TIOCSWINSZ, winsize)
            # execute a new program, replacing the current process
            os.execvpe(exeargv[0], exeargv, host_env)
        # open as binary I/O. buffered and flush method support
        pid = os.fork()
        if pid == 0:
            self.serve_child_terminal(fdm=fdm, read_pipe=r)
        else:
            yield from self.serve_parent_terminal(fdm=fdm, term_id=term_id, write_pipe=w)

    @local
    def serve_parent_terminal(self, fdm, term_id, write_pipe):
        serve_user_input_th = Thread(target=self.serve_user_input, args=(term_id, write_pipe))
        serve_user_input_th.start()
        while True:
            try:
                # Read as much as possible according to max allowed daffi payload in one message.
                b = os.read(fdm, BYTES_CHUNK)
            except OSError:
                break
            yield b

    @local
    def serve_child_terminal(self, fdm, read_pipe):
        pty = os.fdopen(fdm, "wb")
        while True:
            b = read_pipe.recv()
            command, data = b[:1], b[1:]
            if command == Command.STOP:
                pty.write(b"\x04")  # Write Ctrl-D
                pty.flush()
                read_pipe.close()
                os._exit(0)

            elif command == Command.RESIZE:
                height, width = data.decode().split(",")
                winsize = struct.pack("HHHH", int(height), int(width), 0, 0)
                fcntl.ioctl(0, tty.TIOCSWINSZ, winsize)
            else:
                pty.write(data)
                pty.flush()

    @local
    def serve_user_input(self, term_id, write_pipe):
        try:
            for data in self.router_fetcher.write_to_terminal(term_id).get():
                write_pipe.send(data)
                if data == Command.STOP:
                    write_pipe.close()
                    break
        except RemoteStoppedUnexpectedly as e:
            logger.error(str(e))
            write_pipe.send(Command.STOP.value)
            write_pipe.close()



from argparse import Namespace
from daffi import Global
from daffi_terminals.worker.worker import Worker, RouterFetcher


def start_worker(args: Namespace):
    router_fetcher = RouterFetcher()
    Worker(router_fetcher=router_fetcher)
    Global(
        process_name=args.name,
        host=args.rpc_host, port=args.rpc_port,
        ssl_certificate=args.ssl_cert, ssl_key=args.ssl_key,
        on_node_connect=router_fetcher.on_worker_connect_cb,
    ).join()



import os
import uvicorn
import asyncio
import logging
from pathlib import Path
from threading import Event
from dataclasses import dataclass, asdict, field
from contextlib import asynccontextmanager
from queue import Queue
from typing import Dict, Optional
from fastapi import FastAPI, WebSocket, APIRouter
from fastapi.responses import FileResponse
from daffi import Global, FG
from daffi.exceptions import RemoteStoppedUnexpectedly
from daffi.decorators import fetcher, __body_unknown__, local
from daffi.registry import Callback
from fastapi.staticfiles import StaticFiles
from fastapi.websockets import WebSocketDisconnect

STOP_MARKER = b""
ROUTER_ROOT = Path(__file__).parent
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("router")


class NonUniqueNameDetected(Exception):
    ...


@dataclass
class Worker:
    host: str
    mac: str
    process_name: str
    id: str
    active: bool = field(repr=False)

    def serialize(self):
        return asdict(self)


class WebHandler:

    def __init__(self, web_host: str, web_port: int):
        self.web_host = web_host
        self.web_port = web_port

        self.ws_queues: Dict[int, Queue] = dict()
        self.director_sockets: Dict[int, WebSocket] = dict()

        self.workers: Dict[str, Worker] = dict()
        self.disconnect_worker_event = Event()

        self.app = FastAPI(lifespan=self.lifespan)
        self.static = StaticFiles(directory=ROUTER_ROOT / "static")
        self.app.mount("/static", self.static, name="static")

        api_router = APIRouter()
        # Web handlers:
        #     `/`         - index page (render html).
        #     `/director` - main socket. Manage interaction with UI.
        #     `/terminal` - terminal socket. Manage user input from keyboard.
        api_router.add_api_route("/", self.index, methods=["GET"])
        api_router.add_api_websocket_route("/director", self.director)
        api_router.add_api_websocket_route("/terminal", self.terminal)
        self.app.include_router(api_router)

    @fetcher
    @staticmethod
    async def read_from_terminal(term_id: int):
        """Read output data from remote terminal"""
        __body_unknown__(term_id)

    @asynccontextmanager
    async def lifespan(self, *_):
        # Initialize `Global` object before web.
        disconect_worker_obs_task = asyncio.create_task(self.on_worker_disconnect_observer())
        yield
        disconect_worker_obs_task.cancel()
        os._exit(0)

    async def index(self):
        return FileResponse(self.static.directory / "index.html", media_type="text/html")

    async def director(self, websocket: WebSocket):
        """Director socket handler"""
        await websocket.accept()
        director_id = websocket.__hash__()
        self.director_sockets[director_id] = websocket
        await websocket.send_json([w.serialize() for w in self.workers.values()])
        # Read director commands
        async for data in websocket.iter_json():
            if data["command"] == "delete_terminal":
                term_id = data["term_id"]
                self.workers.pop(term_id, None)
                await self.update_workers()
        self.director_sockets.pop(director_id, None)

    async def terminal(self, websocket: WebSocket):
        """Terminal socket handler"""
        await websocket.accept()
        term_id = websocket.__hash__()
        worker_id = websocket.query_params["worker_id"]
        queue = Queue()
        self.ws_queues[term_id] = queue

        read_term_task = asyncio.create_task(
            self.read_terminal_output(term_id=term_id, worker_id=worker_id, websocket=websocket)
        )
        try:
            async for payload in websocket.iter_bytes():
                queue.put_nowait(payload)
        except (WebSocketDisconnect, RuntimeError):
            ...
        queue.put_nowait(STOP_MARKER)
        read_term_task.cancel()

    async def read_terminal_output(self, term_id: int, worker_id: str, websocket: WebSocket):
        """Get output data from remote terminal and write it to websocket"""
        exec_modifier = FG(receiver=worker_id)
        try:
            read_term_iterator = await self.read_from_terminal.call(term_id=term_id, exec_modifier=exec_modifier)
        except Exception as e:
            logger.error(f"unexpected error during read from remote terminal: {e}")
        else:
            try:
                async for out in read_term_iterator.get_async():
                    await websocket.send_bytes(out)
            except RemoteStoppedUnexpectedly:
                if existing_worker := self.workers.get(worker_id):
                    existing_worker.active = False
                    await self.update_workers()
            await websocket.close()

    async def update_workers(self):
        """Send list of connected/disconnected workers through director socket."""
        workers = [w.serialize() for w in self.workers.values()]
        for director_socket in self.director_sockets.values():
            await director_socket.send_json(workers)

    def on_worker_disconnect(self, _, process_name: str):
        if worker := self.workers.get(process_name):
            worker.active = False
            logger.info(f"{worker} has been disconnected.")
            self.disconnect_worker_event.set()
            self.disconnect_worker_event = Event()

    async def on_worker_disconnect_observer(self):
        while True:
            await asyncio.get_running_loop().run_in_executor(None, self.disconnect_worker_event.wait)
            await self.update_workers()

    def run(self):
        uvicorn.run(self.app, port=self.web_port, host=self.web_host)


class Router(Callback):

    def __init__(
            self, rpc_host: str, rpc_port: int, ssl_cert: Optional[str],
            ssl_key: Optional[str], web_handler: WebHandler
    ):
        super().__init__()
        self.name = "TermRouter"
        self.rpc_host = rpc_host
        self.rpc_port = rpc_port
        self.ssl_cert = ssl_cert
        self.ssl_key = ssl_key

        self.web_handler = web_handler
        self.ws_queues = self.web_handler.ws_queues
        self.workers = self.web_handler.workers
        self.director_sockets = self.web_handler.director_sockets

    @local
    def run(self):
        Global(
            host=self.rpc_host, port=self.rpc_port, init_controller=True,
            ssl_certificate=self.ssl_cert, ssl_key=self.ssl_key,
            process_name=self.name, on_node_disconnect=self.web_handler.on_worker_disconnect
        )
        self.web_handler.run()

    def write_to_terminal(self, term_id: int):
        """
        The method involves retrieving information from user input through a websocket.
        It is a remote callback that continuously transfers data from
        a `router` application to a worker process that operates a pty terminal.
        """
        queue = self.ws_queues[term_id]
        while True:
            payload = queue.get()
            yield payload
            if payload == STOP_MARKER:
                del self.ws_queues[term_id]
                break

    async def on_worker_connect(self, host, mac, process_name, id):
        duplicate_name_detected = False
        worker = Worker(host=host, mac=mac, process_name=process_name, id=id, active=True)
        # Check if worker with such `process_name` already exists
        if existing_worker := self.workers.get(process_name):
            if existing_worker.id != worker.id and existing_worker.active:
                duplicate_name_detected = True
                worker = existing_worker
                existing_worker.active = False
        self.workers[process_name] = worker
        logger.info(f"{worker} has been connected.")
        await self.web_handler.update_workers()

        if duplicate_name_detected:
            raise NonUniqueNameDetected(f"Worker with name {process_name!r} already exists and is active."
                                        f" To fix this re-connect existing worker with name {process_name!r}"
                                        f" and make sure all workers have unique names.")
